""" DB connection utility functions """
import subprocess
from functools import lru_cache
import sys
from sqlalchemy import create_engine
from sqlalchemy.engine import Connection
from sqlalchemy.orm import Session
from sqlalchemy.exc import DatabaseError
from langchain_community.vectorstores.pgembedding import CollectionStore
from brevia.settings import get_settings


def connection_string() -> str:
    """ Postgres+pgvector db connection string """
    return get_settings().connection_string()


@lru_cache
def get_engine():
    """Return db engine (using lru_cache)"""
    return create_engine(
            connection_string(),
            pool_size=get_settings().pgvector_pool_size,
        )


def db_connection() -> Connection:
    """ SQLAlchemy db connection """
    return get_engine().connect()


def test_connection() -> bool:
    """ Test db connection with a simple query """
    try:
        with Session(db_connection()) as session:
            (
                session.query(CollectionStore.uuid)
                .limit(1)
                .all()
            )
        return True
    except DatabaseError:
        print('Error performing a simple SQL query', file=sys.stderr)
        return False


def psql_command(
    cmd: str,
) -> bool:
    """Perform command on db using `psql`"""
    conf = get_settings()
    host = conf.pgvector_host
    port = conf.pgvector_port
    database = conf.pgvector_database
    user = conf.pgvector_user
    password = conf.pgvector_password

    cmd = ['psql', '-U', user, '-h', host, '-p', f'{port}', '-c', cmd, database]

    with subprocess.Popen(
        cmd,
        env={'PGPASSWORD': password},
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE
    ) as proc:
        res = proc.communicate()
        output = res[0].decode('UTF-8')
        print(f"psql output: {output}")
        if proc.returncode != 0:
            print('Error launching psql command', file=sys.stderr)
            return False

    print('Psql command executed successfully')
    return True



"""Completion API endpoints"""
from langchain.chains.base import Chain
from fastapi import APIRouter
from brevia.dependencies import (
    get_dependencies,
)
from brevia.completions import CompletionParams, simple_completion_chain
from brevia.callback import token_usage_callback, token_usage, TokensCallbackHandler

router = APIRouter()


class CompletionBody(CompletionParams):
    """ /completion request body """
    text: str
    prompt: dict | None = None
    token_data: bool = False


@router.post('/completion', dependencies=get_dependencies())
async def completion_action(
    completion_body: CompletionBody,
):
    """ /completion endpoint, send a text with a custom prompt and get a completion """

    chain = simple_completion_chain(
        completion_params=CompletionParams(**completion_body.model_dump()),
    )

    return await run_chain(
        chain=chain,
        completion_body=completion_body,
    )


async def run_chain(
    chain: Chain,
    completion_body: CompletionBody,
):
    """Run chain usign async methods and return result"""
    with token_usage_callback() as callb:
        result = await chain.ainvoke({
            'text': completion_body.text
        }, return_only_outputs=True)
    return completion_result(
        result=result,
        callb=callb,
    )


def completion_result(
    result: dict,
    callb: TokensCallbackHandler,
) -> dict:
    """ Handle chat result: save chat history and return answer """
    answer = result['text'].strip(" \n")

    return {
        'completion': answer,
        'usage': token_usage(callb),
    }



"""Utilities to create langchain LLM and Chat Model instances."""
from abc import ABC, abstractmethod
from typing import Any
from langchain_core.embeddings import Embeddings
from langchain_core.language_models import BaseChatModel, BaseLLM
from langchain_community.chat_models.fake import FakeListChatModel
from langchain_community.llms.fake import FakeListLLM
from langchain_community.llms.loading import load_llm_from_config
from langchain_community.embeddings.fake import FakeEmbeddings
from langchain_openai import ChatOpenAI, OpenAIEmbeddings
from openai import OpenAI
from brevia.settings import get_settings
from brevia.utilities.types import load_type


class FakeBreviaLLM(FakeListLLM):
    """Fake LLM for testing purposes."""

    def get_token_ids(self, text: str) -> list[int]:
        """Fake for testing purposes."""
        return [10] * 10


LOREM_IPSUM = """Lorem ipsum dolor sit amet, consectetur adipisici elit,
sed eiusmod tempor incidunt ut labore et dolore magna aliqua."""


def load_llm(config: dict) -> BaseLLM:
    """Load langchain LLM, use Fake LLM in test mode"""
    if test_models_in_use():
        return FakeBreviaLLM(responses=[LOREM_IPSUM] * 10)

    return load_llm_from_config(config=config)


class FakeBreviaChatModel(FakeListChatModel):
    """Fake Chat Model for testing purposes."""

    def get_token_ids(self, text: str) -> list[int]:
        """Fake for testing purposes."""
        return [10] * 10


def load_chatmodel(config: dict) -> BaseChatModel:
    """Load Chat Model from Config Dict."""
    if test_models_in_use():
        return FakeBreviaChatModel(responses=[LOREM_IPSUM] * 10)

    chatmodel_aliases = {
        'openai-chat': ChatOpenAI,
        'fake-list-chat-model': FakeListChatModel,
    }
    model_type = config.pop('_type')
    if model_type in chatmodel_aliases:
        llm_cls = chatmodel_aliases[model_type]
    else:
        llm_cls = load_type(
            model_type,
            BaseChatModel,
            'langchain_community.chat_models',
        )

    return llm_cls(**config)


class BaseAudio(ABC):
    """Base class for Audio transcription"""
    @abstractmethod
    def transcribe(self, file: Any, **kwargs) -> dict:
        """transcribe Audio"""


class FakeAudio(BaseAudio):
    """Test class for Audio transcription"""

    def transcribe(self, file: Any, **kwargs) -> dict:
        """transcribe Audio"""
        return {'text': LOREM_IPSUM}


class AudioOpenAI(BaseAudio):
    """OpenAI Audio transcription"""

    def transcribe(self, file: Any, **kwargs) -> dict:
        """transcribe Audio"""
        client = OpenAI()
        audio_file = open(file, "rb")
        return client.audio.transcriptions.create(
            model="whisper-1",
            file=audio_file,
            **kwargs
        )


def load_audiotranscriber() -> BaseAudio:
    """Load Audio transcriber (only openAI supported for now)."""
    if test_models_in_use():
        return FakeAudio()

    return AudioOpenAI()


def load_embeddings(custom_conf: dict | None = None) -> Embeddings:
    """ Load Embeddings engine """
    settings = get_settings()
    if test_models_in_use():
        return FakeEmbeddings(size=1536)

    config = settings.embeddings.copy() if not custom_conf else custom_conf
    embed_aliases = {
        'openai-embeddings': OpenAIEmbeddings,
        'fake-embeddings': FakeEmbeddings,
    }
    model_type = config.pop('_type')
    if model_type in embed_aliases:
        emb_cls = embed_aliases[model_type]
    else:
        emb_cls = load_type(
            model_type,
            Embeddings,
            'langchain_community.embeddings',
        )

    return emb_cls(**config)


def test_models_in_use() -> bool:
    """Check if test models are in use (via `USE_TEST_MODELS` env var)"""
    return get_settings().use_test_models



"""API endpoints definitions to handle audio input"""
from typing_extensions import Annotated
from fastapi import APIRouter, HTTPException, status, Depends
from pydantic import BaseModel
from brevia.dependencies import get_dependencies
from brevia.chat_history import (
    get_history,
    get_history_sessions,
    history_evaluation,
    ChatHistoryFilter,
)

router = APIRouter()


@router.get(
    '/chat_history',
    dependencies=get_dependencies(json_content_type=False),
    tags=['Chat'],
)
def read_chat_history(filter: Annotated[ChatHistoryFilter, Depends()]):
    """ /chat_history endpoint, read stored chat history """
    return get_history(filter=filter)


@router.get(
    '/chat_history/sessions',
    dependencies=get_dependencies(json_content_type=False),
    tags=['Chat'],
)
def read_chat_history_sessions(filter: Annotated[ChatHistoryFilter, Depends()]):
    """ /chat_history sessions endpoint, read stored chat history sessions"""
    return get_history_sessions(filter=filter)


class EvaluateBody(BaseModel):
    """ Evaluation creation model """
    uuid: str
    user_evaluation: bool
    user_feedback: str | None = None
    metadata: dict | None = None


@router.post(
    '/evaluate',
    status_code=204,
    dependencies=get_dependencies(),
    tags=['Chat'],
)
def evluate_chat_history(body: EvaluateBody):
    """ /evaluate endpoint, save chat history item user evaluation """
    result = history_evaluation(
        history_id=body.uuid,
        user_evaluation=body.user_evaluation,
        user_feedback=body.user_feedback,
        metadata=body.metadata,
    )
    if not result:
        raise HTTPException(
            status.HTTP_404_NOT_FOUND,
            f"Chat history '{body.uuid}' was not found",
        )



"""Question-answering and search functions against a vector database."""
from os import path
from langchain.retrievers.multi_query import MultiQueryRetriever
from langchain.chains.base import Chain
from langchain.chains.conversational_retrieval.base import ConversationalRetrievalChain
from langchain.chains.question_answering import load_qa_chain
from langchain_core.language_models import BaseLanguageModel
from langchain_core.retrievers import BaseRetriever
from langchain_core.vectorstores import VectorStore
from langchain.chains.llm import LLMChain
from langchain_community.vectorstores.pgembedding import CollectionStore
from langchain_community.vectorstores.pgvector import DistanceStrategy, PGVector
from langchain_core.callbacks import BaseCallbackHandler
from langchain_core.documents import Document
from langchain_core.prompts import (
    ChatPromptTemplate,
    HumanMessagePromptTemplate,
    PromptTemplate,
    SystemMessagePromptTemplate,
    load_prompt,
)
from langchain_core.prompts.loading import load_prompt_from_config
from pydantic import BaseModel
from brevia.connection import connection_string
from brevia.collections import single_collection_by_name
from brevia.callback import AsyncLoggingCallbackHandler
from brevia.models import load_chatmodel, load_embeddings
from brevia.settings import get_settings
from brevia.utilities.types import load_type

# system = load_prompt(f'{prompts_path}/qa/default.system.yaml')
# jinja2 template from file was disabled by langchain so, for now
# we load from here
SYSTEM_TEMPLATE = """
                As an AI assistant your task is to provide valuable
                information and support to our users. Answer the question
                as truthfully as possible using the provided context
                between ##Context start## and ##Context end##.
                If the answer is not contained within the provided context,
                say that you are sorry but that you cannot answer
                to that question. Don't try to make up an answer.
                ##Context start## {{context}} ##Context end##
                Answer in {% if lang|length %}{{ lang }}{% else %}
                the same language of the question{% endif %}
                """


def load_qa_prompt(prompts: dict | None) -> ChatPromptTemplate:
    """
        load prompts for RAG-Q/A functions, following openIA system/human/ai pattern:
        SYSTEM_TEMPLATE: is a jinja2 template with language checking from api call.
                         For security reason, the default is loaded from here and
                         overwritten from api call.
        HUMAN_TEMPLATE:  is a normal template, loaded from /prompts/qa and overwritten
                         by api call

    """

    prompts_path = f'{path.dirname(__file__)}/prompts'
    system = PromptTemplate.from_template(SYSTEM_TEMPLATE, template_format="jinja2")
    human = load_prompt(f'{prompts_path}/qa/default.human.yaml')

    if prompts:
        if prompts.get('system'):
            system = load_prompt_from_config(prompts.get('system'))
        if prompts.get('human'):
            human = load_prompt_from_config(prompts.get('human'))

    system_message_prompt = SystemMessagePromptTemplate(prompt=system)
    human_message_prompt = HumanMessagePromptTemplate(prompt=human)

    return ChatPromptTemplate.from_messages(
        [system_message_prompt, human_message_prompt]
    )


def load_condense_prompt(prompts: dict | None) -> ChatPromptTemplate:
    """
        Check if specific few-shot prompt file exists for the collection, if not,
        check for condense prompt file, otherwise, load default condense prompt file.
        few-shot can improve condensing the chat history especially when the chat
        history is about eterogenous topics.
    """
    if prompts:
        if prompts.get('few'):
            return load_prompt_from_config(prompts.get('few'))
        if prompts.get('condense'):
            return load_prompt_from_config(prompts.get('condense'))

    prompts_path = f'{path.dirname(__file__)}/prompts'

    return load_prompt(f'{prompts_path}/qa/default.condense.yaml')


DISTANCE_MAP = {
    'euclidean': DistanceStrategy.EUCLIDEAN,
    'cosine': DistanceStrategy.COSINE,
    'max': DistanceStrategy.MAX_INNER_PRODUCT,
}


class SearchQuery(BaseModel):
    """ Search query items """
    query: str
    collection: str
    docs_num: int | None = None
    distance_strategy_name: str = 'cosine'
    filter: dict[str, str | dict | list] | None = None


def search_vector_qa(
    search: SearchQuery,
) -> list[tuple[Document, float]]:
    """ Perform a similarity search on vector index """
    collection_store = single_collection_by_name(search.collection)
    if not collection_store:
        raise ValueError(f'Collection not found: {search.collection}')
    if search.docs_num is None:
        default_num = get_settings().search_docs_num
        search.docs_num = int(collection_store.cmetadata.get('docs_num', default_num))
    strategy = DISTANCE_MAP.get(search.distance_strategy_name, DistanceStrategy.COSINE)
    embeddings_conf = collection_store.cmetadata.get('embedding', None)
    docsearch = PGVector(
        connection_string=connection_string(),
        embedding_function=load_embeddings(embeddings_conf),
        collection_name=search.collection,
        distance_strategy=strategy,
        use_jsonb=True,
    )

    return docsearch.similarity_search_with_score(
        query=search.query,
        k=search.docs_num,
        filter=search.filter,
    )


class ChatParams(BaseModel):
    """ Q&A basic conversation chain params"""
    docs_num: int | None = None
    streaming: bool = False
    distance_strategy_name: str | None = None
    filter: dict[str, str | dict] | None = None
    source_docs: bool = False
    multiquery: bool = False


def create_custom_retriever(
        store: VectorStore,
        search_kwargs: dict,
        retriever_conf: dict,
) -> BaseRetriever:
    """
        Create a custom retriever from a configuration.
    """
    retriever_name = retriever_conf.pop('retriever', '')
    retriever_class = load_type(retriever_name, BaseRetriever)

    return retriever_class(
        vectorstore=store,
        search_kwargs=search_kwargs,
        **retriever_conf,
    )


def create_default_retriever(
        store: VectorStore,
        search_kwargs: dict,
        llm: BaseLanguageModel,
        multiquery: bool = False,

) -> BaseRetriever:
    """
        Create a default retriever.
        Can be a vector store retriever or a multiquery retriever.
    """
    retriever = store.as_retriever(search_kwargs=search_kwargs)
    if multiquery:
        return MultiQueryRetriever.from_llm(retriever=retriever, llm=llm)

    return retriever


def conversation_chain(
    collection: CollectionStore,
    chat_params: ChatParams,
    answer_callbacks: list[BaseCallbackHandler] | None = None,
    conversation_callbacks: list[BaseCallbackHandler] | None = None,
) -> Chain:
    """
        Return conversation chain for Q/A with embdedded dataset knowledge

        collection: collection store item
        chat_params: basic conversation chain parameters, including:
            docs_num: number of docs to retrieve to create context
                (default 'SEARCH_DOCS_NUM' env var or '4')
            streaming: activate streaming (default False),
            distance_strategy_name: distance strategy to use (default 'cosine')
            filter: optional dictionary of metadata to use as filter (defailt None)
            source_docs: flag to retrieve source docs in response (default True)
        answer_callbacks: callbacks to use in the final LLM answer to enable streaming
            (default empty list)
        conversation_callbacks: callback to handle conversation results
            (default empty list)
        multiquery: flag for activate langchain's multiquery retriver

        can implement "vectordbkwargs" into quest_dict:
            {
                "search_distance": 0.9
            }
    """
    settings = get_settings()
    if chat_params.docs_num is None:
        default_num = settings.search_docs_num
        chat_params.docs_num = int(collection.cmetadata.get('docs_num', default_num))
    if answer_callbacks is None:
        answer_callbacks = []
    if conversation_callbacks is None:
        conversation_callbacks = []

    strategy = DISTANCE_MAP.get(
        chat_params.distance_strategy_name,
        DistanceStrategy.COSINE
    )
    docsearch = PGVector(
        connection_string=connection_string(),
        embedding_function=load_embeddings(collection.cmetadata.get('embedding', None)),
        collection_name=collection.name,
        distance_strategy=strategy,
        use_jsonb=True,
    )

    prompts = collection.cmetadata.get('prompts')
    qa_llm_conf = collection.cmetadata.get(
        'qa_completion_llm',
        settings.qa_completion_llm.copy()
    )
    fup_llm_conf = collection.cmetadata.get(
        'qa_followup_llm',
        settings.qa_followup_llm.copy()
    )

    verbose = settings.verbose_mode

    # LLM to rewrite follow-up question
    fup_llm = load_chatmodel(fup_llm_conf)

    logging_handler = AsyncLoggingCallbackHandler()
    # Create chain for follow-up question using chat history (if present)
    question_generator = LLMChain(
        llm=fup_llm,
        prompt=load_condense_prompt(prompts),
        verbose=verbose,
        callbacks=[logging_handler],
    )

    # Model to use in final prompt
    answer_callbacks.append(logging_handler)
    qa_llm_conf['callbacks'] = answer_callbacks
    qa_llm_conf['streaming'] = chat_params.streaming
    chatllm = load_chatmodel(qa_llm_conf)

    # this chain use "stuff" to elaborate context
    doc_chain = load_qa_chain(
        llm=chatllm,
        prompt=load_qa_prompt(prompts),
        chain_type="stuff",
        verbose=verbose,
        callbacks=[logging_handler],
    )

    # main chain, do all the jobs
    search_kwargs = {'k': chat_params.docs_num, 'filter': chat_params.filter}
    retriever_conf = collection.cmetadata.get(
        'qa_retriever',
        settings.qa_retriever.copy()
    )
    if not retriever_conf:
        retriever = create_default_retriever(
            store=docsearch,
            search_kwargs=search_kwargs,
            llm=chatllm,
            multiquery=chat_params.multiquery,
        )
    else:   # custom retriever
        retriever = create_custom_retriever(docsearch, search_kwargs, retriever_conf)

    conversation_callbacks.append(logging_handler)

    return ConversationalRetrievalChain(
        retriever=retriever,
        combine_docs_chain=doc_chain,
        return_source_documents=chat_params.source_docs,
        question_generator=question_generator,
        callbacks=conversation_callbacks,
        verbose=verbose,
    )



"""API endpoints definitions to handle audio input"""
from typing import Annotated
from fastapi import APIRouter, Response, status, Header
from brevia.dependencies import token_auth
from brevia.connection import test_connection
from brevia.settings import get_settings

router = APIRouter()


def check_authorization(token: str | None, authorization: str | None):
    """Get /status endpoint authorization"""
    if not get_settings().tokens_secret:
        return
    if get_settings().status_token and token == get_settings().status_token:
        return
    if authorization and authorization.startswith('Bearer '):
        token = authorization.removeprefix('Bearer ')
    else:
        token = ''
    token_auth(token=token)


@router.api_route(
    '/status',
    methods=['GET', 'HEAD'],
    tags=['Status'],
)
def api_status(
    response: Response,
    token: str | None = None,
    authorization: Annotated[str | None, Header()] = None
):
    """ /status endpoint, safety check """
    check_authorization(token=token, authorization=authorization)
    db_status = test_connection()
    if not db_status:
        response.status_code = status.HTTP_503_SERVICE_UNAVAILABLE

    return {
        'db_status': 'OK' if db_status else 'KO'
    }



"""Utility commands for applications"""
import json
import sys
from os import getcwd, path
from logging import config
import click
from brevia.alembic import current, upgrade, downgrade
from brevia.index import update_links_documents
from brevia.utilities import files_import, run_service, collections_io
from brevia.tokens import create_token
from brevia.utilities.openapi import brevia_openapi


def init_logging():
    """Initialize logging from optional `log.ini` file"""
    log_ini_path = f'{getcwd()}/log.ini'
    if path.exists(log_ini_path):
        config.fileConfig(log_ini_path)


@click.command()
@click.option("-v", "--verbose", is_flag=True, default=False, help="Verbose mode")
def db_current_cmd(verbose):
    """Display current database revision"""
    current(verbose)


@click.command()
@click.option("-r", "--revision", default="head", help="Revision target")
def db_upgrade_cmd(revision):
    """Upgrade to a later database revision"""
    upgrade(revision)


@click.command()
@click.option("-r", "--revision", required=True, help="Revision target")
def db_downgrade_cmd(revision):
    """Revert to a previous database revision"""
    downgrade(revision)


@click.command()
@click.option("-f", "--file-path", required=True, help="File or folder path")
@click.option("-c", "--collection", required=True, help="Collection name")
@click.option("-o", "--options", required=False, help="Loader options in JSON format")
def import_file(file_path: str, collection: str, options: str = ''):
    """Add file or folder content to collection"""
    # pass loader options via JSON string
    kwargs = {} if not options else json.loads(options)
    num = files_import.index_file_folder(
        file_path=file_path,
        collection=collection,
        **kwargs
    )
    print(
        f"""Collection '{collection}'
        updated from '{file_path}' with {num} documents."""
    )


# A `test_service.yml` in the root folder is needed.
# `service` and `payload` keys are required with this structure:
#
# service: my_module.MyService
# output_path: path/to/output.txt
# payload:
#   file_path: /path/to/file
#   param1: value1
#   param2: value2
@click.command()
@click.option("-n", "--num", default=1, help="Number of attempts")
@click.option(
    "-f",
    "--file-path",
    default=f'{getcwd()}/test_service.yml',
    help="yaml file path"
)
def run_test_service(num: int = 1, file_path: str = f'{getcwd()}/test_service.yml'):
    """Service test"""
    # allow module import from current working directory
    sys.path.append(getcwd())
    run_service.run_service_from_yaml(file_path=file_path, num_attempts=num)


@click.command()
@click.option("-c", "--collection", required=True, help="Collection name")
@click.option(
    "-f",
    "--folder-path",
    required=True,
    help="Folder output path"
)
def export_collection(folder_path: str, collection: str):
    """Export a collection to CSV postgres files."""
    collections_io.export_collection_data(
        folder_path=folder_path,
        collection=collection
    )


@click.command()
@click.option("-c", "--collection", required=True, help="Collection name")
@click.option(
    "-f",
    "--folder-path",
    required=True,
    help="Folder input path"
)
def import_collection(folder_path: str, collection: str):
    """Import a collection from a CSV postgres files."""
    collections_io.import_collection_data(
        folder_path=folder_path,
        collection=collection
    )


@click.command()
@click.option("-u", "--user", default="brevia", help="Token user name")
@click.option("-d", "--duration", default=60, help="Token duration in minutes")
def create_access_token(user: str, duration: int):
    """Create an access token """
    token = create_token(user=user, duration=duration)
    print(token)


@click.command()
@click.option(
    "-f",
    "--file-path",
    default=f'{getcwd()}/pyproject.toml',
    required=True,
    help="Path to a pyproject.toml file"
)
@click.option(
    "-o",
    "--output",
    default='./openapi.json',
    required=True,
    help="Path of the generated OpenAPI JSON file"
)
def create_openapi(file_path: str, output: str):
    """Create an openapi metadata file"""
    metadata = brevia_openapi(py_proj_path=file_path)
    with open(output, 'w') as f:  # pylint: disable=unspecified-encoding
        json.dump(metadata, f)


@click.command()
@click.option("-c", "--collection", required=True, help="Collection name")
def update_collection_links(collection: str):
    """Update index documents of a collection having `"type": "links"` in metadata."""
    init_logging()
    num = update_links_documents(collection_name=collection)
    print(f'Updated {num} links documents. Done!')



"""API endpoints to Brevia configuration"""
from typing import Any
from fastapi import APIRouter, HTTPException, status
from pydantic import ValidationError
from brevia.settings import (
    Settings,
    get_configurable_keys,
    get_settings,
    reset_db_conf,
    update_db_conf,
)
from brevia.dependencies import get_dependencies
from brevia.connection import db_connection

router = APIRouter()


@router.api_route(
    '/config',
    methods=['GET', 'HEAD'],
    dependencies=get_dependencies(json_content_type=False),
    tags=['Config'],
)
def get_config():
    """ /config endpoint, read Brevia configuration """
    return get_settings().model_dump()


@router.api_route(
    '/config/schema',
    methods=['GET', 'HEAD'],
    dependencies=get_dependencies(json_content_type=False),
    tags=['Config'],
)
def get_config_schema():
    """ /config/schema endpoint, read Brevia configuration settings keys """
    return get_settings().get_configurable_schema()


def check_keys(keys: list[str]):
    """Check if all keys are configurable"""
    not_configurable = [
        key for key in keys if key not in get_configurable_keys()
    ]
    if not_configurable:
        raise HTTPException(
            status.HTTP_400_BAD_REQUEST,
            f'There are not configurable settings: {",".join(not_configurable)}',
        )


@router.api_route(
    '/config',
    methods=['POST'],
    dependencies=get_dependencies(),
    tags=['Config'],
)
def save_config(config: dict[str, Any]):
    """POST /config endpoint, save Brevia configuration """
    check_keys(config.keys())

    try:
        Settings(**config)
    except ValidationError as exc:
        raise HTTPException(
            status.HTTP_400_BAD_REQUEST,
            f'Invalid configuration: {exc}',
        )

    return update_db_conf(db_connection(), config)


@router.api_route(
    '/config/reset',
    methods=['POST'],
    dependencies=get_dependencies(),
    tags=['Config'],
)
def reset_config(keys: list[str]):
    """POST /config/reset endpoint, reset to default a list of settings """
    check_keys(keys)

    return reset_db_conf(db_connection(), keys)



"""Utility functions to import files or folders."""
import os
from typing import List, Any
from langchain_core.documents import Document
from brevia.index import add_document
from brevia.load_file import read


def index_file_folder(
    file_path: str,
    collection: str,
    **kwargs: Any,
) -> int:
    """ Load documents in collection index from a file or folder """
    if not os.path.exists(file_path):
        raise FileNotFoundError(f"File {file_path} does not exist.")

    chunks_num = 0
    docs = load_file_folder_documents(file_path=file_path, **kwargs)
    for doc in docs:
        chunks_num += add_document(
            document=doc,
            collection_name=collection,
        )

    return chunks_num


def load_file_folder_documents(file_path: str, **kwargs: Any) -> List[Document]:
    """Load documents from file or folder"""
    if os.path.isdir(file_path):
        docs = []
        for file in os.listdir(file_path):
            docs.append(Document(
                page_content=read(file_path=f'{file_path}/{file}', **kwargs),
                metadata={'type': 'files', 'path': file_path},
            ))
        return docs

    doc = Document(
        page_content=read(file_path=file_path, **kwargs),
        metadata={'type': 'files', 'path': file_path},
    )

    return [doc]



"""Add brevia app routers."""
from fastapi import FastAPI
from brevia.routers import (
    analyze_router,
    index_router,
    collections_router,
    chat_history_router,
    audio_router,
    jobs_router,
    qa_router,
    status_router,
    completion_router,
    config_router,
)


def add_routers(app: FastAPI) -> None:
    """Add brevia APP routers to FastAPI instance"""
    app.include_router(index_router.router)
    app.include_router(analyze_router.router)
    app.include_router(collections_router.router)
    app.include_router(audio_router.router)
    app.include_router(jobs_router.router)
    app.include_router(chat_history_router.router)
    app.include_router(qa_router.router)
    app.include_router(status_router.router)
    app.include_router(completion_router.router)
    app.include_router(config_router.router)



"""Collections handling functions"""
from langchain_community.vectorstores.pgembedding import CollectionStore
from sqlalchemy.orm import Session
from brevia import connection
from brevia.utilities.uuid import is_valid_uuid


def collections_info(collection: str | None = None):
    """ Retrieve collections data """
    filter_collection = CollectionStore.name == collection
    if collection is None:
        filter_collection = CollectionStore.name is not None

    with Session(connection.db_connection()) as session:
        query = (
            session.query(
                CollectionStore.uuid,
                CollectionStore.name,
                CollectionStore.cmetadata,
            )
            .filter(filter_collection)
            .limit(100)
        )

        result = [u._asdict() for u in query.all()]
        session.close()

    return result


def collection_name_exists(name: str) -> bool:
    """ Check if a collection name exists already """
    store = single_collection_by_name(name)

    return store is not None


def collection_exists(uuid: str) -> bool:
    """ Check if a collection uuid exists """
    store = single_collection(uuid)

    return store is not None


def single_collection(uuid: str) -> (CollectionStore | None):
    """ Get single collection by UUID """
    if not is_valid_uuid(uuid):
        return None
    with Session(connection.db_connection()) as session:
        return session.get(CollectionStore, uuid)


def single_collection_by_name(name: str) -> (CollectionStore | None):
    """ Get single collection by name"""
    with Session(connection.db_connection()) as session:
        return CollectionStore.get_by_name(session=session, name=name)


def create_collection(
    name: str,
    cmetadata: dict,
) -> CollectionStore:
    """ Create single collection """
    with Session(connection.db_connection()) as session:
        collection_store = CollectionStore(
            name=name,
            cmetadata=cmetadata,
        )
        session.expire_on_commit = False
        session.add(collection_store)
        session.commit()

        return collection_store


def update_collection(
    uuid: str,
    name: str,
    cmetadata: dict,
):
    """ Update single collection """
    with Session(connection.db_connection()) as session:
        collection = session.get(CollectionStore, uuid)
        collection.name = name
        collection.cmetadata = cmetadata
        session.add(collection)
        session.commit()


def delete_collection(
    uuid: str,
):
    """ Delete single collection """
    with Session(connection.db_connection()) as session:
        collection = session.get(CollectionStore, uuid)
        session.delete(collection)
        session.commit()



"""Brevia Middleware classes"""
from importlib.metadata import version
from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint
from starlette.requests import Request
from starlette.responses import Response
from starlette.types import ASGIApp


class VersionHeaderMiddleware(BaseHTTPMiddleware):
    """Middleware to add version headers to response"""
    def __init__(
        self,
        app: ASGIApp,
        api_version: str = '',
        api_name: str = '',
    ) -> None:
        super().__init__(app)
        self.api_version = api_version
        self.api_name = api_name

    async def dispatch(
        self, request: Request, call_next: RequestResponseEndpoint
    ) -> Response:
        """Add version header to response"""
        response = await call_next(request)
        response.headers['X-Brevia-Version'] = version('Brevia')
        # Add API version and name headers
        # about the custom API created with Brevia
        if self.api_name:
            response.headers['X-API-Name'] = self.api_name
        if self.api_version:
            response.headers['X-API-Version'] = self.api_version

        return response



"""Base service and some basic implementations"""
from abc import ABC, abstractmethod
from os import unlink
from brevia.callback import token_usage_callback
from brevia.tasks.text_analysis import RefineTextAnalysisTask
from brevia.analysis import summarize
from brevia.load_file import read


class BaseService(ABC):
    """Base class for services"""

    def run(self, payload: dict) -> dict:
        """Run a service using a payload"""
        if not self.validate(payload):
            raise ValueError(f'Invalid service payload - {payload}')
        return self.execute(payload)

    @abstractmethod
    def execute(self, payload: dict) -> dict:
        """Execute service logic using payload"""

    @abstractmethod
    def validate(self, payload: dict) -> bool:
        """Validate service payload"""


class SummarizeTextService(BaseService):
    """Service to perform summarization from text input"""

    def execute(self, payload: dict):
        """Service logic"""
        token_data = payload.pop('token_data')
        with token_usage_callback() as callb:
            result = summarize(**payload)

        return {
            'output': result['output_text'],
            'token_data': None if not token_data else callb.__dict__
        }

    def validate(self, payload: dict):
        """Payload validation"""
        if not payload.get('text'):
            return False
        return True


class SummarizeFileService(BaseService):
    """Service to perform summarization from file input"""

    def execute(self, payload: dict):
        """Service logic"""
        keys = ['file_path', 'chain_type', 'initial_prompt',
                'iteration_prompt', 'token_data']
        args = dict(filter(lambda item: item[0] in keys, payload.items()))

        return self.summarize_from_file(**args)

    def validate(self, payload: dict):
        """Payload validation"""
        if not payload.get('file_path'):
            return False
        return True

    def summarize_from_file(
        self,
        file_path: str,
        chain_type: str | None = None,
        initial_prompt: dict | None = None,
        iteration_prompt: dict | None = None,
        token_data: bool = False,
    ) -> dict:
        """
        Perform summarization of a temporary PDF or TXT file.
        File is removed after summarization.
        """
        try:
            text = read(file_path=file_path)
        finally:
            unlink(file_path)  # Delete the temp file

        if not text:
            raise ValueError('Empty text')

        with token_usage_callback() as callb:
            result = summarize(
                text,
                chain_type=chain_type,
                initial_prompt=initial_prompt,
                iteration_prompt=iteration_prompt
            )

        return {
            'output': result['output_text'],
            'token_data': None if not token_data else callb.__dict__
        }


class RefineTextAnalysisService(BaseService):
    """Service to perform summarization from text input"""

    def execute(self, payload: dict):
        """Service logic"""
        analysis = RefineTextAnalysisTask(
            file_path=payload['file_path'],
            prompts=payload['prompts'],
            llm_conf=payload.get('llm_conf'),
            text_options=payload.get('text_options')
        )
        with token_usage_callback() as callb:
            result = analysis.perform_task()
        token_data = callb.__dict__
        token_data.pop('_lock', None)

        return {
            'output': result['output_text'],
            'token_data': token_data
        }

    def validate(self, payload: dict):
        """Payload validation"""
        if not payload.get('file_path') or not payload.get('prompts'):
            return False
        prompts = payload['prompts']
        if 'initial_prompt' not in prompts or 'refine_prompt' not in prompts:
            return False

        return True


class FakeService(BaseService):
    """Fake class for services testing"""

    def execute(self, payload: dict) -> dict:
        return {'output': 'ok'}

    def validate(self, payload: dict) -> bool:
        return True



"""Simple completion functions"""
from langchain.chains.base import Chain
from langchain.chains.llm import LLMChain
from langchain_core.prompts.loading import load_prompt_from_config
from pydantic import BaseModel
from brevia.models import load_chatmodel
from brevia.settings import get_settings


class CompletionParams(BaseModel):
    """ Q&A basic conversation chain params """
    prompt: dict | None = None


def load_custom_prompt(prompt: dict | None):
    """ Load custom prompt """
    return load_prompt_from_config(prompt)


def simple_completion_chain(
    completion_params: CompletionParams,
) -> Chain:
    """
        Return simple completion chain for a generic input text

        completion_params: basic completion params including:
            prompt: custom prompt for execute simple completion commands
    """

    settings = get_settings()
    llm_conf = settings.qa_completion_llm.copy()
    comp_llm = load_chatmodel(llm_conf)
    verbose = settings.verbose_mode
    # Create chain for follow-up question using chat history (if present)
    completion_llm = LLMChain(
        llm=comp_llm,
        prompt=load_custom_prompt(completion_params.prompt),
        verbose=verbose,
    )

    return completion_llm



"""chat history evaluation

Revision ID: b34017752e45
Revises: 1d927e0cbc61
Create Date: 2024-01-17 11:10:55.127272

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'b34017752e45'
down_revision = '1d927e0cbc61'
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.add_column(
        'chat_history',
        sa.Column(
            'user_evaluation',
            sa.BOOLEAN(),
            nullable=True,
            comment='User evaluation as good (True) or bad (False)',
        ),
    )
    op.add_column(
        'chat_history',
        sa.Column(
            'user_feedback',
            sa.String(),
            nullable=True,
            comment='User textual feedback on the evaluation',
        ),
    )
    op.add_column(
        'chat_history',
        sa.Column(
            'chat_source',
            sa.String(),
            nullable=True,
            comment='Generic string to identify chat source (e.g. application name)',
        ),
    )


def downgrade() -> None:
    op.drop_column('chat_history', 'user_evaluation')
    op.drop_column('chat_history', 'user_feedback')
    op.drop_column('chat_history', 'chat_source')



"""API endpoints definitions to handle async jobs"""
from fastapi import APIRouter, HTTPException, status
from brevia.dependencies import get_dependencies
from brevia import async_jobs

router = APIRouter()


@router.get(
    '/jobs/{uuid}',
    dependencies=get_dependencies(json_content_type=False),
    tags=['Analysis'],
)
async def read_analysis_job(uuid: str):
    """
    Read details of a single analisys Job via its UUID
    """
    job = async_jobs.single_job(uuid)
    if job is None:
        raise HTTPException(
            status.HTTP_404_NOT_FOUND,
            f"Job '{uuid}' was not found",
        )

    return job



"""Functions to perform summarization & document analysis"""
from langchain.chains.summarize import load_summarize_chain
from langchain_core.documents import Document
from langchain_core.language_models import BaseChatModel
from langchain_core.prompts.loading import load_prompt_from_config
from langchain_text_splitters import TokenTextSplitter
from brevia.callback import LoggingCallbackHandler
from brevia.models import load_chatmodel
from brevia.settings import get_settings


def load_stuff_prompts(prompts: dict | None) -> dict:
    """Load custom prompts for the 'stuff' summarization chain.
    Summarization chain is the simplest one and have only one prompt

    Args:
        prompts (dict | None): Optional prompts for customization.
        It should be a dictionary with the following keys:
        - 'initial_prompt' (dict): The prompt to use for the initial summarization.

    Returns:
        dict: Custom prompts for 'stuff' summarization chain.
    """
    if prompts.get('initial_prompt'):
        return {
            'prompt': load_prompt_from_config(prompts.get('initial_prompt'))
        }

    return {}


def load_map_prompts(prompts: dict | None) -> dict:
    """Load custom prompts for the 'map_reduce' summarization chain.
    The 'map_reduce' summarization chain utilizes a more sophisticated algorithm that
    merges multiple prompts to create a summary. It divides the text into text chunks
    and applies the initial prompt to each of them. Afterward, it uses the iteration
    prompt to combine all the result into a final summary.

    Args:
        prompts (dict | None): Optional prompts for customization.
        It should be a dictionary with the following keys:
        - 'initial_prompt' (dict): The prompt to use for the chunks summarization.
        - 'iteration_prompt' (dict | None): (Optional) The prompt to use for summarizing
            all the pieces.

    Returns:
        dict: Custom prompts for 'map_reduce' summarization chain.
    """
    prompts_data = {}
    if prompts.get('initial_prompt'):
        map_prompt = load_prompt_from_config(prompts.get('initial_prompt'))
        prompts_data['map_prompt'] = map_prompt
        prompts_data['combine_prompt'] = map_prompt
        if prompts.get('iteration_prompt'):
            combine_prompt = load_prompt_from_config(prompts.get('iteration_prompt'))
            prompts_data['combine_prompt'] = combine_prompt

    return prompts_data


def load_refine_prompts(prompts: dict | None) -> dict:
    """Load custom prompts for the 'refine' summarization chain.

    The 'refine' summarization chain breaks the document into chunks pieces, then
    concentrates on making the summary better by refining it.
    Starts using the initial prompt for the first chunk then uses the iteration prompt
    over the remaining chunks to further enhance the summary.

    Args:
        prompts (dict | None): Optional prompts for customization.
        It should be a dictionary with the following keys:
        - 'initial_prompt' (dict): The prompt to use for the first chunk summarization.
        - 'iteration_prompt' (dict | None): (Optional) The prompt to use for increase
        information over the chunk pieces.

    Returns:
        dict: Custom prompts for 'refine' summarization chain.
    """

    prompts_data = {}
    if prompts.get('initial_prompt'):
        question_prompt = load_prompt_from_config(prompts.get('initial_prompt'))
        prompts_data['question_prompt'] = question_prompt
        if prompts.get('iteration_prompt'):
            refine_prompt = load_prompt_from_config(prompts.get('iteration_prompt'))
            prompts_data['refine_prompt'] = refine_prompt

    return prompts_data


def get_summarize_llm() -> BaseChatModel:
    """Get a summarization language model for text summarization.

    This function returns a language model for use in the summarization process.
    The model's configuration is defined by environment variables.

    Returns:
        BaseChatModel: A language model suitable for text summarization.

    """
    model = get_settings().summarize_llm.copy()  # pylint: disable=no-member
    model['callbacks'] = [LoggingCallbackHandler()]

    return load_chatmodel(model)


def summarize(
    text: str,
    chain_type: str | None = None,
    initial_prompt: dict | None = None,
    iteration_prompt: dict | None = None
) -> str:
    """Perform summarizing for a given text.
    This function takes a text as input and generates a summary using
    langchain summarization chain main alghoritms (stuff, map_reduce, refine).

    Args:
        text: The input text that you want to summarize.
        chain_type: The main langchain summarization chain type should be one of
            "stuff", "map_reduce", and "refine".
            if not providerd stuff is used by default
        initial_prompt: Optional custom prompt to be used in the selected langchain
            chain type to replace the main chain prompt defaults.
        iteration_prompt: Optional custom prompts to be used in the selected
            langchain chain type to replace the second chain promopt defaults.

    Returns:
        str: The generated summary of the input text.

    Raises:
        ValueError: If an unsupported summarization chain type is specified.
    """

    summarize_chain_map = {
        'stuff': load_stuff_prompts,
        'map_reduce': load_map_prompts,
        'refine': load_refine_prompts
    }
    settings = get_settings()
    chain_type = chain_type or settings.summ_default_chain
    if chain_type not in summarize_chain_map:
        raise ValueError(
            f"Got unsupported chain type: {chain_type}. "
            f"Should be one of {list(summarize_chain_map.keys())}"
        )

    logging_handler = LoggingCallbackHandler()
    kwargs = {
        'llm': get_summarize_llm(),
        'chain_type': chain_type,
        'verbose': settings.verbose_mode,
        'callbacks': [logging_handler],
    }

    # load chain_type specific prompts:
    prompts_args = summarize_chain_map[chain_type]({
        'initial_prompt': initial_prompt,
        'iteration_prompt': iteration_prompt
    })

    chain = load_summarize_chain(**kwargs, **prompts_args)
    text_splitter = TokenTextSplitter(
        chunk_size=settings.summ_token_splitter,
        chunk_overlap=settings.summ_token_overlap
    )
    texts = text_splitter.split_text(text)
    docs = [Document(page_content=t) for t in texts]

    return chain.invoke({'input_documents': docs}, return_only_outputs=True)



"""Utility functions to import/export collections using CSV postgres files."""
from os import path
from brevia import connection, collections


def export_collection_data(
    folder_path: str,
    collection: str,
):
    """Export collection data using `psql`"""
    if not collections.collection_name_exists(collection):
        raise ValueError(f"Collection '{collection}' was not found")

    csv_file_collection = f"{folder_path}/{collection}-collection.csv"
    csv_file_embedding = f"{folder_path}/{collection}-embedding.csv"
    if path.exists(csv_file_collection):
        raise ValueError(f"CSV file {csv_file_collection} already exists, exiting")
    if path.exists(csv_file_embedding):
        raise ValueError(f"CSV file {csv_file_embedding} already exists, exiting")

    select = f"SELECT * FROM langchain_pg_collection WHERE name = '{collection}'"
    copy = f"\\copy ({select}) TO '{csv_file_collection}' WITH DELIMITER ',' CSV HEADER"
    res = connection.psql_command(cmd=copy)
    if not res:
        return

    print(f"Collection record '{collection}' exported to '{csv_file_collection}'")

    sub = f"SELECT uuid from langchain_pg_collection WHERE name = '{collection}'"
    select = f"SELECT * FROM langchain_pg_embedding WHERE collection_id = ({sub})"
    copy = f"\\copy ({select}) TO '{csv_file_embedding}' WITH DELIMITER ',' CSV HEADER"
    res = connection.psql_command(cmd=copy)
    if not res:
        return

    print(f"Collection embedding for '{collection}' exported to '{csv_file_embedding}'")


def import_collection_data(
    folder_path: str,
    collection: str,
):
    """Import collection data using `psql`"""
    if collections.collection_name_exists(collection):
        raise ValueError(f"Collection '{collection}' already exists, exiting")

    csv_file_collection = f"{folder_path}/{collection}-collection.csv"
    csv_file_embedding = f"{folder_path}/{collection}-embedding.csv"
    print(f"Importing from {csv_file_collection} and {csv_file_embedding}")
    if not path.exists(csv_file_collection):
        raise ValueError(f"CSV file {csv_file_collection} not found, exiting")
    if not path.exists(csv_file_embedding):
        raise ValueError(f"CSV file {csv_file_embedding} not found, exiting")

    table = 'langchain_pg_collection'
    copy = f"\\copy {table} FROM '{csv_file_collection}' WITH DELIMITER ',' CSV HEADER"
    res = connection.psql_command(cmd=copy)
    if not res:
        return

    table = 'langchain_pg_embedding'
    copy = f"\\copy {table} FROM '{csv_file_embedding}' WITH DELIMITER ',' CSV HEADER"
    res = connection.psql_command(cmd=copy)
    if not res:
        return

    print(f"Collection '{collection}' imported")



"""JWT Utility methods"""
from datetime import datetime, timedelta
from jose import jwt, JWTError
from brevia.settings import get_settings

ALGORITHM = 'HS256'


def create_token(user: str, duration: int) -> str:
    """Create an access token """
    # expire time of the token
    expire = datetime.utcnow() + timedelta(minutes=duration)
    to_encode = {
        'user': user,
        'exp': expire,
    }
    secret = get_settings().tokens_secret

    return jwt.encode(to_encode, secret, algorithm=ALGORITHM)


def verify_token(token: str):
    """
        Try to decode the token, it will
        raise a JWTError if the token is not correct
    """
    settings = get_settings()
    secret = settings.tokens_secret
    payload = jwt.decode(token, secret, algorithms=[ALGORITHM])
    # basic check: presence of `user` key, must be a non empty string
    if 'user' not in payload:
        raise JWTError('Bad payload format')
    user = payload['user']
    if not isinstance(user, str) or len(user.strip()) == 0:
        raise JWTError('Bad payload format')
    # further check: user must one of `TOKENS_USERS` env var (if set)
    valid_users = settings.tokens_users
    if not valid_users:
        return
    if user not in valid_users.split(','):
        raise JWTError('Bad payload format')



"""API endpoints definitions for indexing actions"""
from typing import Annotated
from os import path
import json
import re
import logging
from pydantic import BaseModel
from fastapi import APIRouter, HTTPException, Request, status, UploadFile, Form
from langchain_community.vectorstores.pgembedding import CollectionStore
from langchain_core.documents import Document
from brevia.dependencies import (
    get_dependencies,
    save_upload_file_tmp,
    check_collection_uuid,
)
from brevia import index, collections, load_file

router = APIRouter()


class IndexBody(BaseModel):
    """ /index request body """
    content: str
    collection_id: str
    document_id: str
    metadata: dict = {}


@router.post(
    '/index',
    status_code=204,
    dependencies=get_dependencies(),
    tags=['Index'],
)
def index_document(item: IndexBody):
    """ Add single document to collection index """
    collection = load_collection(collection_id=item.collection_id)
    # remove same document if already indexed
    index.remove_document(
        collection_id=item.collection_id,
        document_id=item.document_id,
    )
    index.add_document(
        document=Document(page_content=item.content, metadata=item.metadata),
        collection_name=collection.name,
        document_id=item.document_id,
    )


def load_collection(collection_id: str) -> CollectionStore:
    """ Load collection by ID and throw 404 if not found"""
    collection = collections.single_collection(collection_id)
    if collection is None:
        raise HTTPException(
            status.HTTP_404_NOT_FOUND,
            f"Collection id '{collection_id}' was not found",
        )
    return collection


@router.post(
    '/index/upload',
    status_code=204,
    dependencies=get_dependencies(json_content_type=False),
    tags=['Index'],
)
def upload_and_index(
    file: UploadFile,
    collection_id: Annotated[str, Form()],
    document_id: Annotated[str, Form()],
    metadata: Annotated[str | None, Form()] = None,
    options: Annotated[str | None, Form()] = None,
):
    """
    Upload a PDF file and perform index on a collection
    """
    collection = load_collection(collection_id=collection_id)
    log = logging.getLogger(__name__)
    log.info("Uploaded '%s' - %s - %s", file.filename, file.content_type, file.size)
    log.info("Collection '%s' - document %s", collection.name, document_id)
    if file.content_type not in ['application/pdf', 'text/plain']:
        raise HTTPException(
            status.HTTP_400_BAD_REQUEST,
            f'Unsupported file content type "{file.content_type}"',
        )
    tmp_path = save_upload_file_tmp(file)
    if metadata is not None:
        metadata = json.loads(metadata)
    else:
        metadata = {'source': path.basename(tmp_path)}

    # remove same document if already indexed
    index.remove_document(
        collection_id=collection_id,
        document_id=document_id,
    )
    read_options = {} if options is None else json.loads(options)
    text = load_file.read(file_path=tmp_path, **read_options)
    index.add_document(
        document=Document(
            page_content=text,
            metadata=metadata,
        ),
        collection_name=collection.name,
        document_id=document_id,
    )


class IndexLink(BaseModel):
    """ /index/link request body """
    link: str  # link to a webpage
    collection_id: str
    document_id: str
    metadata: dict = {}
    options: dict = {}


@router.post(
    '/index/link',
    status_code=204,
    dependencies=get_dependencies(json_content_type=False),
    tags=['Index'],
)
def parse_link_and_index(item: IndexLink):
    """
    Add a web page content to a collection index
    """
    collection = load_collection(collection_id=item.collection_id)
    log = logging.getLogger(__name__)
    log.info(
        "Adding link '%s' to collection '%s' / document '%s'",
        item.link,
        collection.name,
        item.document_id
    )

    text = load_file.read_html_url(url=item.link, **item.options)
    if not text:
        return
    # remove same document if already indexed
    index.remove_document(
        collection_id=item.collection_id,
        document_id=item.document_id,
    )
    index.add_document(
        document=Document(
            page_content=text,
            metadata=item.metadata,
        ),
        collection_name=collection.name,
        document_id=item.document_id,
    )


@router.delete(
    '/index/{collection_id}/{document_id}',
    status_code=204,
    dependencies=get_dependencies(json_content_type=False),
    tags=['Index'],
)
def remove_document(collection_id: str, document_id: str):
    """ Remove document from collection index """
    index.remove_document(
        collection_id=collection_id,
        document_id=document_id,
    )


def read_filter(request: Request) -> dict:
    """Read metadata filter dict from query string"""
    result = {}
    regexp = r'filter\[(.*?)\]'
    for key in request.query_params.keys():
        if re.match(regexp, key):
            f_key = re.search(regexp, key).group(1)
            result[f_key] = request.query_params.get(key)

    return result


@router.get(
    '/index/{collection_id}',
    dependencies=get_dependencies(json_content_type=False),
    tags=['Index'],
)
def index_docs(
    collection_id: str,
    request: Request,
    page: int = 1,
    page_size: int = 50
):
    """ Read collection documents with metadata filter """
    load_collection(collection_id=collection_id)
    return index.collection_documents(
        collection_id=collection_id,
        filter=read_filter(request=request),
        page=page,
        page_size=page_size,
    )


@router.get(
    '/index/{collection_id}/documents_metadata',
    dependencies=get_dependencies(json_content_type=False),
    tags=['Index'],
)
def index_docs_metadata(collection_id: str, request: Request):
    """ Read collection documents metadata"""
    load_collection(collection_id=collection_id)
    return index.documents_metadata(
        collection_id=collection_id,
        filter=read_filter(request=request),
        document_id=request.query_params.get('document_id'),
    )


@router.get(
    '/index/{collection_id}/{document_id}',
    dependencies=get_dependencies(json_content_type=False),
    tags=['Index'],
)
def read_document(collection_id: str, document_id: str):
    """ Read single document from collection index """
    return index.read_document(
        collection_id=collection_id,
        document_id=document_id,
    )


class IndexMetaBody(BaseModel):
    """ /index/metadata request body """
    collection_id: str
    document_id: str
    metadata: dict = {}


@router.post('/index/metadata', status_code=204, dependencies=get_dependencies())
def index_metadata_document(item: IndexMetaBody):
    """ Update metadata of a single document in a collection"""
    check_collection_uuid(item.collection_id)
    index.update_metadata(
        collection_id=item.collection_id,
        document_id=item.document_id,
        metadata=item.metadata,
    )



"""Base class for text analysis services"""
from langchain.docstore.document import Document
from langchain_core.callbacks import Callbacks
from langchain.chains.combine_documents.refine import RefineDocumentsChain
from langchain.chains.llm import LLMChain
from langchain_core.language_models import BaseLanguageModel
from langchain_core.prompts.loading import load_prompt, load_prompt_from_config
from brevia.callback import LoggingCallbackHandler
from brevia.load_file import read
from brevia.index import split_document
from brevia.models import load_chatmodel
from brevia.settings import get_settings
from brevia.tasks.base import BaseAnalysisTask


class BaseTextAnalysisTask(BaseAnalysisTask):
    """Base class for text analysis tasks"""

    def text_documents(
            self,
            file_path: str,
            options: dict | None = None
    ) -> list[Document]:
        """Load text from file and split into documents"""
        text = read(file_path=file_path)
        options = options or {}

        return split_document(Document(page_content=text), collection_meta=options)


class RefineTextAnalysisTask(BaseTextAnalysisTask):
    """Text analysis using refine chain"""
    def __init__(
        self,
        file_path: str,
        prompts: dict | None = None,
        llm_conf: dict | None = None,
        text_options: dict | None = None,
    ):
        self.file_path = file_path
        self.llm_conf = llm_conf
        self.text_options = text_options
        self.load_analysis_prompts(prompts)

    def perform_task(self):
        """Service logic"""
        text_documents = self.text_documents(
                self.file_path, options=self.text_options)

        result = self.run_refine_chain(text_documents)

        return {
            'input_documents': [{
                'page_content': doc.page_content,
                'metadata': doc.metadata
            } for doc in text_documents
            ],
            'output_text': result.get('output_text')
        }

    def load_analysis_prompts(self, prompts: dict | None = None):
        """Load analysis prompts"""
        if not prompts:
            raise ValueError('Prompts dictionary must be provided.')
        missing = {'initial_prompt', 'refine_prompt'} - prompts.keys()
        if missing:
            raise ValueError(f'Missing required prompts: {", ".join(missing)}')

        for key, value in prompts.items():
            if isinstance(value, dict):
                prompts[key] = load_prompt_from_config(value)
            else:
                prompts[key] = load_prompt(self.prompt_path(value))

        self.prompts = prompts

    def run_refine_chain(
        self,
        pages: list[Document],
    ) -> dict[str, str]:
        """Run analysis with refine chain"""
        logging_handler = LoggingCallbackHandler()
        llm_conf = self.llm_conf or get_settings().summarize_llm.copy()
        llm_conf['callbacks'] = [logging_handler]
        llm_text = load_chatmodel(llm_conf)

        # Using refine algorithm with modified prompts
        # for informations extraction on multiple pages
        chain = self.load_refine_chain(
            llm=llm_text,
            callbacks=[logging_handler],
        )

        return chain.invoke({'input_documents': pages}, return_only_outputs=True)

    def load_refine_chain(
        self,
        llm: BaseLanguageModel,
        callbacks: Callbacks,
        verbose: bool = True,
    ) -> RefineDocumentsChain:
        """Load refine chain"""
        initial_prompt = self.prompts.get('initial_prompt')
        refine_prompt = self.prompts.get('refine_prompt')

        return RefineDocumentsChain(
            initial_llm_chain=LLMChain(llm=llm, prompt=initial_prompt, verbose=verbose),
            refine_llm_chain=LLMChain(llm=llm, prompt=refine_prompt, verbose=verbose),
            document_variable_name='text',
            initial_response_name='existing_answer',
            verbose=verbose,
            callbacks=callbacks,
        )
